/*****************************************************************************************
 *
 *  Copyright 2006+ by Cieszak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,

 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#include "game.h"
#include "menu.h"

// this global variable keeps track of the game, even if you quit the whole app
struct game *game;

/**************************************************************************************/

#define OMSG_TIMER   (0x1f04)
#define OMSG_GAME_TIMER   (0x1f05)
static const ULONG timer=OMSG_TIMER;
static const ULONG game_timer=OMSG_GAME_TIMER;
static const ULONG key_timer=MSG_KEYPRESS;
extern const struct bitmap bmap_player;
extern const struct bitmap bmap_fixed_wall;
extern const struct bitmap bmap_breakable_wall;
extern const struct bitmap bmap_bomb;
extern const struct bitmap bmap_fire_1;
extern const struct bitmap bmap_fire_2;
extern const struct bitmap bmap_fire_3;
extern const struct bitmap bmap_fire_4;
extern const struct bitmap bmap_exit;
extern const struct bitmap bmap_monster;
extern const struct bitmap bmap_monster2;
extern const struct bitmap bmap_bonus_life;
extern const struct bitmap bmap_bonus_cash;
extern const struct bitmap bmap_bonus_drive;
extern const struct bitmap bmap_bonus_power;
extern const struct bitmap bmap_bonus_bomb;

extern const ringtone_t tone_tick;
extern const ringtone_t tone_boom;
extern const ringtone_t tone_newmonster;
extern const ringtone_t tone_monsterkilled;
extern const ringtone_t tone_foundexit;
extern const ringtone_t tone_gameover;
extern const ringtone_t tone_start;
extern const ringtone_t tone_end;
extern const ringtone_t tone_bonus;
extern const ringtone_t tone_foundbonus;
extern const ringtone_t tone_cheat;
//used for error signaling
extern const ringtone_t tone_error;

//cheats
const unsigned char cheat_life[] = {5, 4, 3, 3, 0};
const unsigned char cheat_power[] = {7, 6, 9, 3, 7};
const unsigned char cheat_bomb[] = {2, 6, 6, 2, 0};
const unsigned char cheat_drive[] = {3, 7, 4, 8, 3};


/*****************************************************************************/

/* TODO: list implementation for bombs, fires and monster tables */


/* shows error msg and play error sound */
/* example usage:
   char *text=malloc(100);
   sprintf(text, "read_map: Mapp bigger than accepted\n");
   error(text);
*/

void error(char* text)
{
   if(!game->error)
   {
      dprintf(text);
      free(text);
      bm_play_sound(&tone_error);
      game->error=1;
   }
}

/* frees the game structure */
void bm_free_game(void)
{
   if(game)
   {
      free(game->maps);
      free(game);
      game=NULL;
   }
}

/* checks player start position */
void bm_get_player_start(signed int *pos_x, signed int *pos_y)
{
   dprintf("bm_get_player_start game->map_width=%d  game->map_height=%d \n", game->map_width, game->map_width);
   unsigned char map_x, map_y;
   for ( map_x = 0; map_x < game->map_width; map_x++ )
   {
      for ( map_y = 0; map_y < game->map_height; map_y++ )
      {
         if(game->map[map_y][map_x] == MAP_PLAYER)
         {
            bm_map_to_pos(map_x, map_y, pos_x, pos_y);
            game->map[map_y][map_x] = MAP_NONE;
            dprintf(" player start position = %d %d\n", map_x, map_y);
            return;
         }
      }
   }
   dprintf("player not found \n");
   bm_map_to_pos(1, 1, pos_x, pos_y);
}

/* calculate map coord */
void inline bm_pos_to_map(signed int pos_x, signed int pos_y, signed char *map_x, signed char *map_y)
{
   *map_x = pos_x/6;
   *map_y = pos_y/6;
}


/* calculate map coord */
void inline bm_map_to_pos(signed char map_x, signed char map_y, signed int *pos_x, signed int *pos_y)
{
   *pos_x = map_x*6;
   *pos_y = map_y*6;
}

/* calculate distance between global points */
unsigned int bm_distance(signed int x1, signed int y1, signed int x2, signed int y2)
{
   return sqrt(pow(x1-x2, 2) + pow(y1 - y2, 2));
}

// return what is on given coord
unsigned char bm_check_wall(signed char map_x, signed char map_y)
{
   if(game->map_width>=MAX_MAP_WIDTH)
   {
      char *text=malloc(100);
      sprintf(text, "game->map_width>=MAX_MAP_WIDTH\n");
      error(text);
   }

   if(game->map_height>=MAX_MAP_HEIGHT)
   {
      char *text=malloc(100);
      sprintf(text, "game->map_height>=MAX_MAP_HEIGHT\n");
      error(text);
   }

   if ( map_x < 0 || map_y < 0 || map_x >= game->map_width || map_y >= game->map_height )
      return MAP_NONE;
   else
      return game->map[map_y][map_x];
}

// return: 0 no bomb, 1 bomb
unsigned char bm_check_bomb(signed char map_x, signed char map_y, unsigned char *bomb_id)
{
   unsigned char c;
   for ( c = 0; c < game->bomb_count; c++ )
   if(game->bombs[c].state)
   {
      if ((game->bombs[c].map_x == map_x) && (game->bombs[c].map_y == map_y))
      {
         if(bomb_id)
            *bomb_id = c;
         return 1;
      }
   }

   return 0;
}

// return: 0 no fire, 1 fire
unsigned char bm_check_fire(signed char map_x, signed char map_y)
{
   unsigned char c;
   for ( c = 0; c < game->bomb_count * ( game->bomb_power * 4 + 1 ); c++ )
   if (game->fires[c].state)
   {
      if (game->fires[c].map_x == map_x && game->fires[c].map_y == map_y)
         return 1;
   }

   return 0;
}

// return: 0 no monster, 1 monster
unsigned char bm_check_monster(signed char map_x, signed char map_y, unsigned char *monster_id )
{
   unsigned char c;
   for ( c = 0; c < MAX_MONSTERS; c++ )
      if(game->monsters[c].state)
      {
         signed char monster_map_x, monster_map_y;
         /* check all four monster "corners" */
         bm_pos_to_map(game->monsters[c].x, game->monsters[c].y, &monster_map_x, &monster_map_y );
         if ((map_x == monster_map_x) && (map_y == monster_map_y))
         {
            if(monster_id)
               *monster_id=c;
            return 1;
         }
         bm_pos_to_map(game->monsters[c].x+5, game->monsters[c].y, &monster_map_x, &monster_map_y );
         if ((map_x == monster_map_x) && (map_y == monster_map_y))
         {
            if(monster_id)
               *monster_id=c;
            return 1;
         }
         bm_pos_to_map(game->monsters[c].x, game->monsters[c].y+5, &monster_map_x, &monster_map_y );
         if ((map_x == monster_map_x) && (map_y == monster_map_y))
         {
            if(monster_id)
               *monster_id=c;
            return 1;
         }
         bm_pos_to_map(game->monsters[c].x+5, game->monsters[c].y+5, &monster_map_x, &monster_map_y );
         if ((map_x == monster_map_x) && (map_y == monster_map_y))
         {
            if(monster_id)
               *monster_id=c;
            return 1;
         }
      }

   return 0;
}

unsigned char bm_check_monsters_die()
{
   unsigned char c;
   for ( c = 0; c < MAX_MONSTERS; c++ )
      if(game->monsters[c].state)
            return 0;
   return 1;
}

void bm_calculate_player_pos()
{
   signed int x = game->x;
   signed int y = game->y;
   signed char map_x, map_y;
   switch(game->key)
   {
      case KEY_2:
         y-=PLAYER_STEP_VALUE;
         break;
      case KEY_4:
         x-=PLAYER_STEP_VALUE;
         break;
      case KEY_6:
         x+=PLAYER_STEP_VALUE;
         break;
      case KEY_8:
         y+=PLAYER_STEP_VALUE;
         break;
    }

   unsigned char wall;
   //left upper player's "corner"
   bm_pos_to_map( x, y, &map_x, &map_y );
   wall=bm_check_wall(map_x, map_y);
   if(wall!=MAP_NONE)
   {
      if(wall==MAP_EXIT)
      {
         if(bm_check_monsters_die())
               bm_next_level();
      }
      else
      if(wall == MAP_BONUS_CASH)
      {
         dprintf("Bonus: cash\n");
         game->score += 50;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_POWER)
      {
         dprintf("Bonus: power\n");
         game->bomb_power++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_BOMB)
      {
         dprintf("Bonus: bomb\n");
         game->bomb_count++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_LIFE)
      {
         dprintf("Bonus: life\n");
         game->lives++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_DRIVE)
      {
         dprintf("Bonus: manual drive\n");
         game->manual_driving = 1;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
         return;
   }
   //if there is a bomb...
   if(bm_check_bomb(map_x, map_y, NULL))
   {
      //.. and player isn't on it
      signed char player_map_x, player_map_y;
      bm_pos_to_map(game->x+2, game->y+2, &player_map_x, &player_map_y);
      if(!bm_check_bomb(player_map_x, player_map_y, NULL))
         return;
   }

   //left lower player's "corner"
   bm_pos_to_map( x, y+5, &map_x, &map_y );
   wall=bm_check_wall(map_x, map_y );
   if(wall!=MAP_NONE)
   {
      if(wall==MAP_EXIT)
      {
         if(bm_check_monsters_die())
               bm_next_level();
      }
      else
      if(wall == MAP_BONUS_CASH)
      {
         dprintf("Bonus: cash\n");
         game->score += 50;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_POWER)
      {
         dprintf("Bonus: power\n");
         game->bomb_power++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_BOMB)
      {
         dprintf("Bonus: bomb\n");
         game->bomb_count++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_LIFE)
      {
         dprintf("Bonus: life\n");
         game->lives++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_DRIVE)
      {
         dprintf("Bonus: manual drive\n");
         game->manual_driving = 1;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
         return;
   }
   //if there is a bomb...
   if(bm_check_bomb(map_x, map_y, NULL))
   {
      //.. and player isn't on it
      signed char player_map_x, player_map_y;
      bm_pos_to_map(game->x+2, game->y+2, &player_map_x, &player_map_y);
      if(!bm_check_bomb(player_map_x, player_map_y, NULL))
         return;
   }

   //right upper player's "corner"
   bm_pos_to_map( x+5, y, &map_x, &map_y );
   wall=bm_check_wall(map_x, map_y );
   if(wall!=MAP_NONE)
   {
      if(wall==MAP_EXIT)
      {
         if(bm_check_monsters_die())
               bm_next_level();
      }
      else
      if(wall == MAP_BONUS_CASH)
      {
         dprintf("Bonus: cash\n");
         game->score += 50;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_POWER)
      {
         dprintf("Bonus: power\n");
         game->bomb_power++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_BOMB)
      {
         dprintf("Bonus: bomb\n");
         game->bomb_count++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_LIFE)
      {
         dprintf("Bonus: life\n");
         game->lives++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_DRIVE)
      {
         dprintf("Bonus: manual drive\n");
         game->manual_driving = 1;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
         return;
   }
   //if there is a bomb...
   if(bm_check_bomb(map_x, map_y, NULL))
   {
      //.. and player isn't on it
      signed char player_map_x, player_map_y;
      bm_pos_to_map(game->x+2, game->y+2, &player_map_x, &player_map_y);
      if(!bm_check_bomb(player_map_x, player_map_y, NULL))
         return;
   }

   //right lower player's "corner"
   bm_pos_to_map( x+5, y+5, &map_x, &map_y );
   wall=bm_check_wall(map_x, map_y );
   if(wall!=MAP_NONE)
   {
      if(wall==MAP_EXIT)
      {
         if(bm_check_monsters_die())
               bm_next_level();
      }
      else
      if(wall == MAP_BONUS_CASH)
      {
         dprintf("Bonus: cash\n");
         game->score += 50;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_POWER)
      {
         dprintf("Bonus: power\n");
         game->bomb_power++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_BOMB)
      {
         dprintf("Bonus: bomb\n");
         game->bomb_count++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_LIFE)
      {
         dprintf("Bonus: life\n");
         game->lives++;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
      if(wall == MAP_BONUS_DRIVE)
      {
         dprintf("Bonus: manual drive\n");
         game->manual_driving = 1;
         bm_play_sound(&tone_bonus);
         game->map[map_y][map_x] = MAP_NONE;
      }
      else
         return;
   }
   //if there is a bomb...
   if(bm_check_bomb(map_x, map_y, NULL))
   {
      //.. and player isn't on it
      signed char player_map_x, player_map_y;
      bm_pos_to_map(game->x+2, game->y+2, &player_map_x, &player_map_y);
      if(!bm_check_bomb(player_map_x, player_map_y, NULL))
         return;
   }

   game->x = x;
   game->y = y;
}

void bm_draw()
{
   /* position */
   signed int pos_x, pos_y;
   signed char obj_map_x, obj_map_y;


   signed char first_map_x_to_draw, first_map_y_to_draw, player_map_x, player_map_y,  a_pos_map_x, a_pos_map_y, x_offset=0, y_offset=0;

   #ifdef ENABLE_FADE
      //fade
      if(game->fade_out>1)
      {
         game->fade_out-=1;
         if(game->fade_out<1)
            game->fade_out=1;
         bm_play_sound(&tone_tick);
      }
      else
      if(game->fade_out==1)
      {
         dprintf("game ->fadeout==1\n");
         game->fade_out=0;
         if (game->gameover)
         {
            bm_play_sound(&tone_gameover);
            send_message(MSG_REQUEST_QUIT);
            return;
         }
         else
            bm_reload_level(game->map_id);
      }

      if(game->fade_in>0)
      {
         game->fade_in-=1;
         if(game->fade_in<0)
            game->fade_in=0;
         bm_play_sound(&tone_tick);
      }
   #endif

   unsigned c;

   bm_pos_to_map(game->x, game->y, &player_map_x, &player_map_y);

   x_offset = game->x%6;
   y_offset = game->y%6 ;

   first_map_x_to_draw = player_map_x - SCREEN_WIDTH/2;
   first_map_y_to_draw = player_map_y - SCREEN_HEIGHT/2;

   freegfx_begin(true);
   for(obj_map_x = 1; obj_map_x < SCREEN_WIDTH-1; obj_map_x++)
   {
      for(obj_map_y = 1; obj_map_y < SCREEN_HEIGHT-1; obj_map_y++)
      {
         a_pos_map_x = first_map_x_to_draw + obj_map_x;
         a_pos_map_y = first_map_y_to_draw + obj_map_y;
         /* fade */
         #ifdef ENABLE_FADE
            if(game->fade_in>0)
               if(bm_distance(a_pos_map_x, a_pos_map_y, player_map_x, player_map_y)>=(FADE_LENGTH-game->fade_in))
               {
                  continue;
               }
            if(game->fade_out>0)
               if(bm_distance(a_pos_map_x, a_pos_map_y, player_map_x, player_map_y)>=game->fade_out)
               {
                  continue;
               }
         #endif
         // drawing walls
         switch (bm_check_wall(a_pos_map_x, a_pos_map_y ) )
         {
            case MAP_FIXED:
               {
                  #ifndef BITMAP_ONLY
                     char fix[6][6]={" XXXX ", "XXXXXX", "XXXXXX", "XXXXXX", "XXXXXX", " XXXX "};
                     signed char scr_x, scr_y;
                     for(scr_x=0; scr_x<6; scr_x++)
                        for(scr_y=0; scr_y<6; scr_y++)
                           draw_pixel(obj_map_x * 6 - x_offset+scr_x, obj_map_y * 6 - y_offset+scr_y, (fix[scr_y][scr_x]=='X'?DM_SET:DM_CLEAR));
                  #else
                     own_draw_bitmap(&bmap_fixed_wall,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
                  #endif
               }
               break;
            case MAP_BREAKEBLE:
            case MAP_HIDDENEXIT:
            case MAP_HIDDENBONUS:
               own_draw_bitmap(&bmap_breakable_wall,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            case MAP_EXIT:
               own_draw_bitmap(&bmap_exit,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            /* bonuses */
            case MAP_BONUS_CASH:
               own_draw_bitmap(&bmap_bonus_cash,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            case MAP_BONUS_DRIVE:
               own_draw_bitmap(&bmap_bonus_drive,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            case MAP_BONUS_LIFE:
               own_draw_bitmap(&bmap_bonus_life,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            case MAP_BONUS_POWER:
               own_draw_bitmap(&bmap_bonus_power,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            case MAP_BONUS_BOMB:
               own_draw_bitmap(&bmap_bonus_bomb,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
               break;
            default:
               break;
         }
         // drawing bombs
         for ( c = 0; c < game->bomb_count; c++ )
         if (game->bombs[c].state)
            if (game->bombs[c].map_x == a_pos_map_x && game->bombs[c].map_y == a_pos_map_y)
               own_draw_bitmap(&bmap_bomb,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);

         /* draw player before fires :) */
         if ( ( player_map_x == a_pos_map_x ) && ( player_map_y == a_pos_map_y ) )
            own_draw_bitmap(&bmap_player,obj_map_x * 6,obj_map_y * 6,DM_SET);

         /* drawing monsters */
         for ( c = 0; c < MAX_MONSTERS; c++ )
         if(game->monsters[c].state)
         {
            signed char map_x, map_y;
            bm_pos_to_map(game->monsters[c].x, game->monsters[c].y, &map_x, &map_y);
            if (map_x == a_pos_map_x && map_y == a_pos_map_y)
            {
               signed char x_off = game->monsters[c].x%6;
               signed char y_off = game->monsters[c].y%6 ;

               switch (game->monsters[c].type)
               {
                  case 1:
                     own_draw_bitmap(&bmap_monster,obj_map_x * 6 - x_offset +x_off,obj_map_y * 6 - y_offset +y_off,DM_SET);
                     break;
                  case 2:
                     own_draw_bitmap(&bmap_monster2,obj_map_x * 6 - x_offset +x_off,obj_map_y * 6 - y_offset +y_off,DM_SET);
                     break;
               }
            }
         }

         // drawing fires
         for ( c = 0; c < game->bomb_count * ( game->bomb_power * 4 + 1 ); c++ )
         if(game->fires[c].state)
         {
            if (game->fires[c].map_x == a_pos_map_x && game->fires[c].map_y == a_pos_map_y )
               switch (game->fires[c].state )
               {
                  case 1:
                     own_draw_bitmap(&bmap_fire_1,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
                     break;
                  case 2:
                     own_draw_bitmap(&bmap_fire_2,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
                     break;
                  case 3:
                     own_draw_bitmap(&bmap_fire_3,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
                     break;
                  case 4:
                     own_draw_bitmap(&bmap_fire_4,obj_map_x * 6 - x_offset,obj_map_y * 6 - y_offset,DM_SET);
                     break;
                }
         }
      }
   }
   
   own_draw_box(0, 0, screen_info.width - 1, 6, DM_CLEAR);
   own_draw_line(0, 6, screen_info.width - 1, 6, 1, DM_SET);
   //level
   char buf[10];
   ftoa(buf, game->level);
   own_print_string((const nokstr_t*)"LEV ", 0, 5, own_get_font("console/4"), DM_SET);
   own_print_string((const nokstr_t*)buf, 16, 5, own_get_font("console/4"), DM_SET);
   //score
   ftoa(buf, game->score);
   unsigned char len=strlen(buf)*4;
   own_print_string((const nokstr_t*)"SCORE ", screen_info.width-len-24, 5, own_get_font("console/4"), DM_SET);
   own_print_string((const nokstr_t*)buf, screen_info.width-len, 5, own_get_font("console/4"), DM_SET);
   //lives
   own_draw_box(0, screen_info.height - 6, screen_info.width - 1, screen_info.height - 1, DM_CLEAR);
   own_draw_line(0, screen_info.height - 7, screen_info.width - 1, screen_info.height - 7, 1, DM_SET);
   for(pos_x = 0; pos_x < game->lives; pos_x++)
      own_draw_bitmap(&bmap_player, pos_x*6, screen_info.height - 6, DM_SET);
   //timer
   ftoa(buf, game->timer);
   len = strlen(buf)*4;
   own_print_string((const nokstr_t*)buf, screen_info.width-len, screen_info.height - 1, own_get_font("console/4"), DM_SET);

   #ifdef ENABLE_FADE
      if(game->fade_in>0)
   #endif
   //map name
   {
      if(strlen(game->map_name)>0)
      {
         len = own_get_string_width((const nokstr_t *)(game->map_name),own_get_font("console/4"));
         pos_x=(screen_info.width-len)/2;
         pos_y=(screen_info.height-8);
         own_draw_box(pos_x, pos_y, len+2, 8, DM_CLEAR);
         own_draw_line(pos_x, pos_y, pos_x, screen_info.height-1, 1, DM_SET);
         own_draw_line(pos_x+len+2, pos_y, pos_x+len+2, screen_info.height-1, 1, DM_SET);
         own_draw_line(pos_x, pos_y, pos_x+len+2, pos_y, 1, DM_SET);
         own_print_string((const nokstr_t *)(game->map_name),pos_x+2, screen_info.height-1, own_get_font("console/4"), DM_SET);
      }
   }

   freegfx_end(false);
}

void bm_add_new_fires(signed char map_x, signed char map_y)
{
   unsigned char c, direction = 0;
   signed char map_x_2_check, map_y_2_check;
   unsigned int i = 0;
   for (direction=0; direction < 5; direction++ )
   {
      map_x_2_check = map_x;
      map_y_2_check = map_y;
      for ( c = 0; c < game->bomb_power; c++ )
      {
         //get first free fire
         while (game->fires[i].state)
            i++;

         switch ( direction )
         {
            case DIR_UP:
               map_y_2_check--;
               break;
            case DIR_LEFT:
               map_x_2_check--;
               break;
            case DIR_DOWN:
               map_x_2_check++;
               break;
            case DIR_RIGHT:
               map_y_2_check++;
               break;
            case DIR_NONE:
               break;
         }

         //if there is a bomb, blow up!
         unsigned char bomb_id;
         if(bm_check_bomb(map_x_2_check, map_y_2_check, &bomb_id))
         {
            game->bombs[bomb_id].state=BOMB_TIME-1;
         }
         else
         if(bm_check_wall(map_x_2_check, map_y_2_check)!= MAP_FIXED &&
           !bm_check_fire(map_x_2_check, map_y_2_check))
         {
            game->fires[i].state = 1;
            game->fires[i].map_x = map_x_2_check;
            game->fires[i].map_y = map_y_2_check;
            game->fires[i].check=0;
            //if breakable wall (or hidden exit, hidden bonus), bomb will destroy only next piece without further
            unsigned char wall=bm_check_wall(map_x_2_check, map_y_2_check);
            if(wall == MAP_BREAKEBLE || wall == MAP_HIDDENEXIT || wall == MAP_HIDDENBONUS)
               break;
         }
         else
            break;
      }
   }
}

void bm_calculate_fires()
{
   unsigned int i;
   for(i=0; i<game->bomb_count*(game->bomb_power*4+1); i++)
   if (game->fires[i].state)
   {
      game->fires[i].state++;
      if(game->fires[i].state >= 4)
      {
         switch (game->map[game->fires[i].map_y][game->fires[i].map_x])
         {
            case MAP_HIDDENEXIT:
               /* hidden exit */
               bm_play_sound(&tone_foundexit);
               game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_EXIT;
               break;

            case MAP_HIDDENBONUS:
               bm_play_sound(&tone_foundbonus);
               /* hidden bonus */
               switch (rand()%5)
               {
                  case 0:
                     game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_BONUS_CASH;
                     break;
                  case 1:
                     game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_BONUS_DRIVE;
                     break;
                  case 2:
                     game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_BONUS_LIFE;
                     break;
                  case 3:
                     game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_BONUS_POWER;
                     break;
                  case 4:
                     game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_BONUS_BOMB;
                     break;
               }
               break;

            case MAP_EXIT:
            case MAP_BONUS_CASH:
            case MAP_BONUS_DRIVE:
            case MAP_BONUS_LIFE:
            case MAP_BONUS_POWER:
            case MAP_BONUS_BOMB:
               /* exit or bonus (create random monsters)*/
               bm_play_sound(&tone_newmonster);
               int j;
               for(j=0; j<3; j++)
               {
                  unsigned char count=0;
                  /* find first free monster entry */
                  while((game->monsters[count].state!=0) && (count < MAX_MONSTERS))
                     count ++;
                  if((game->monsters[count].state==0) && (count < MAX_MONSTERS))
                  {
                     game->monsters[count].type=2;
                     game->monsters[count].state=1;
                     game->monsters[count].direction=rand()%4;
                     bm_map_to_pos(game->fires[i].map_x, game->fires[i].map_y, &game->monsters[count].x, &game->monsters[count].y);
                     count++;
                  }
               }
               break;

            case MAP_BREAKEBLE:
               /* breakable wall */
               game->map[game->fires[i].map_y][game->fires[i].map_x] = MAP_NONE;
               break;
         }
         //kill monster
         unsigned char monster_id;
         if(bm_check_monster(game->fires[i].map_x, game->fires[i].map_y, &monster_id))
         {
            bm_play_sound(&tone_monsterkilled);
            game->monsters[monster_id].state=0;
            game->score += game->monsters[monster_id].type*5;
         }

         //kamikaze :D
         signed int x = game->x;
         signed int y = game->y;
         signed char map_x, map_y;
         bm_pos_to_map(x,y, &map_x, &map_y);
         if (game->fires[i].map_y==map_y && game->fires[i].map_x== map_x)
            bm_kill_player();
         game->fires[i].state = 0;
      }
   }
}

void bm_calculate_bombs()
{
   unsigned char c;
   signed char player_map_x, player_map_y;

   bm_pos_to_map( game->x, game->y, &player_map_x, &player_map_y );

   if (game->key == KEY_5)
   {
      if (!bm_check_bomb(player_map_x, player_map_y, NULL))
         for ( c = 0; c < game->bomb_count; c++ )
            if(!game->bombs[c].state)
            {
               game->bombs[c].map_x = player_map_x;
               game->bombs[c].map_y = player_map_y;
               if(game->manual_driving)
               {
                  /* bombs.state is used for indicate bomb order because timer is unused*/
                  int i, maxid = 0;
                  for(i = 0; i < game->bomb_count; i++)
                  {
                     if(game->bombs[i].state)
                        if(maxid < game->bombs[i].state)
                           maxid = game->bombs[i].state;
                  }
                  maxid++;
                  game->bombs[c].state = maxid;
               }
               else
               {
                  game->bombs[c].state = 1;
               }
               bm_play_sound(&tone_tick);
               break;
            }
   }
   
   if(game->manual_driving)
   {
      if (game->key == KEY_0)
      {
         /* find any used bomb */
         for ( c = 0; c < game->bomb_count; c++ )
         {
            if (game->bombs[c].state)
               break;
         }
         
         unsigned char first = c;
         unsigned char min = game->bombs[first].state;
         dprintf("First used bomb is %d (state = %d)\n", first, min);
         /* find first bomb */
         for ( c = 0; c < game->bomb_count; c++ )
         {
            if (game->bombs[c].state)
            {
               if(min > game->bombs[c].state)
               {
                  min = game->bombs[c].state;
                  first = c;
                  dprintf("Found bomb before previous [%d state = %d]\n", first, min);
               }
            }
         }

         /* blow first bomb */
         if(game->bombs[first].state)
         {
            game->bombs[first].state = 0;
            bm_add_new_fires (game->bombs[first].map_x, game->bombs[first].map_y);
            bm_play_sound(&tone_boom);
         }
      }
   }
   else
   {
      for ( c = 0; c < game->bomb_count; c++ )
      {
         if (game->bombs[c].state)
         {
            game->bombs[c].state++;
            if (game->bombs[c].state >= BOMB_TIME)
            {
               game->bombs[c].state = 0;
               bm_add_new_fires (game->bombs[c].map_x, game->bombs[c].map_y);
               bm_play_sound(&tone_boom);
            }
         }
      }
   }
}

void bm_start_timer()
{
   own_timer_start(&timer, TT_MESSAGE, 20);
}

void bm_start_game_timer()
{
   own_timer_start(&game_timer, TT_MESSAGE, TICKS_PER_SECOND);
}

void bm_stop_timer()
{
   own_timer_abort(&timer);
}

void bm_stop_game_timer()
{
   own_timer_abort(&game_timer);
}

void bm_play_sound(const ringtone_t *tone)
{
   if(own_get_game_setting(GS_SOUNDS))
      own_play_tone(tone,PTM_WARNING_TONE);
}

void bm_read_map()
{
   const unsigned char templ[MAX_MAP_HEIGHT][MAX_MAP_WIDTH] = {
     "FFFFFFFFFFFFFFF",
     "F             F",
     "F F F F F F F F",
     "F             F",
     "F F F F F F F F",
     "F             F",
     "F F F F F F F F",
     "F             F",
     "F F F F F F F F",
     "F             F",
     "F F F F F F F F",
     "F             F",
     "F F F F F F F F",
     "F             F",
     "FFFFFFFFFFFFFFF",
     };

   if(game->maps_count==0)
   {
      dprintf("Generating random map\n");
      /* copy template */
      int i;
      for(i=0; i<MAX_MAP_HEIGHT; i++)
         memcpy(game->map[i], templ[i], MAX_MAP_WIDTH);
      game->map_width=15;
      game->map_height=15;
      /* put player */
      game->map[1][1] = MAP_PLAYER;
      game->map[2][1] = MAP_RESERVED;
      game->map[1][2] = MAP_RESERVED;
      /* random breakable walls */
      unsigned int map_x,map_y;
      for(i=0; i<25; i++)
      {
         do
         {
            map_x=rand()%game->map_width;
            map_y=rand()%game->map_height;
         }
         while(bm_check_wall(map_x, map_y)!=MAP_NONE);
         game->map[map_y][map_x]=MAP_BREAKEBLE;
         dprintf("Added breakable on %d x %d\n", map_x, map_y);
      }
      /* random monsters */
      for(i=0; i<3; i++)
      {
         do
         {
            map_x=rand()%game->map_width;
            map_y=rand()%game->map_height;
         }
         while(bm_check_wall(map_x, map_y)!=MAP_NONE);
         game->map[map_y][map_x]=MAP_MONSTER;
         dprintf("Added monster on %d x %d\n", map_x, map_y);
      }
      /* exit */
      do
      {
         map_x=rand()%game->map_width;
         map_y=rand()%game->map_height;
      }
      while(bm_check_wall(map_x, map_y)!=MAP_NONE);
      game->map[map_y][map_x]=MAP_HIDDENEXIT;
      dprintf("Added exit on %d x %d\n", map_x, map_y);
      /* bonus */
      do
      {
         map_x=rand()%game->map_width;
         map_y=rand()%game->map_height;
      }
      while(bm_check_wall(map_x, map_y)!=MAP_NONE);
      game->map[map_y][map_x] = MAP_HIDDENBONUS;
      dprintf("Added bonus on %d x %d\n", map_x, map_y);
      /* replace MAP_RESERVED to MAP_NONE */
      for(map_x = 0; map_x < game->map_width; map_x++)
      {
         for(map_y = 0; map_y < game->map_height; map_y++)
         {
            if(game->map[map_y][map_x] == MAP_RESERVED)
               game->map[map_y][map_x] = MAP_NONE;
         }
      }
   }
   else //load map from file
   {
      if(!game->map_id)
         game->map_id=game->maps[rand()%game->maps_count];
      memset(game->map_name,0,20);
      game->map_width=0;
      game->map_height=0;
      unsigned char x=0, y=0;
      int fd=tfsopen(tfspath(game->map_id), TFS_RDONLY);
      //allocating read buffer
      char buffer[300];
      while(!tfseof(fd))
      {
         int bytes=tfsread(fd, buffer, 300);
         int i, nexty=0;
         for(i=0; i<bytes; i++)
         {
            if(buffer[i]<32) //\n \r or smth
            {
               if(game->map_width==0)
                  game->map_width=x;
               if(nexty==0)
               {
                  y++;
                  x=0;
                  nexty=1;
               }
            }
            else
            {
               nexty=0;
               if(buffer[i]=='@')//map name
               {
                  if(game->map_height==0)
                     game->map_height=y;
                  x=0;
               }
               else if(game->map_height!=0)//continue map name
               {
                  game->map_name[x]=buffer[i];
                  x++;
               }
               else //map
               if((x<MAX_MAP_WIDTH)&&(y<MAX_MAP_HEIGHT))
               {
                  game->map[y][x]=buffer[i];
                  x++;
               }
               else
               {
                  char *text=malloc(100);
                  sprintf(text, "read_map: Mapp bigger than accepted\n");
                  error(text);
               }
            }
         }
      }
      //if map hasn't name
      if(game->map_height==0)
         game->map_height=y;
      /* validate size values */
      if(game->map_width>=MAX_MAP_WIDTH)
         game->map_width=MAX_MAP_WIDTH-1;

      if(game->map_height>=MAX_MAP_HEIGHT)
         game->map_height=MAX_MAP_HEIGHT-1;

      tfsclose(fd);
      int i;
      for(i=0; i<MAX_MAP_HEIGHT; i++)
         dprintf("%s", game->map[i]);

   }
}

void bm_next_level()
{
   game->level++;
   game->map_id = NULL;
   bm_reload_level();
}

void bm_reload_level()
{
   bm_read_map();
   bm_get_player_start (&(game->x), &(game->y) );

   game->key=0xFF;
   unsigned char  i;
   for ( i = 0; i < game->bomb_count; i++ )
      game->bombs[i].state = 0;
   for ( i = 0; i < game->bomb_count*(game->bomb_power*4+1); i++)
      game->fires[i].state = 0;
   for ( i = 0; i < MAX_MONSTERS; i++ )
      game->monsters[i].state = 0;

   /* creating monsters */
   unsigned char count=0;
   while(bm_get_next_monster_start(&(game->monsters[count].x), &(game->monsters[count].y)))
   {
      game->monsters[count].type=1;
      game->monsters[count].state=1;
      game->monsters[count].direction=rand()%4;
      count++;
   }
   #ifdef ENABLE_FADE
      //preparing fade in
      game->fade_in = FADE_LENGTH;
   #endif
   //start timer;
   game->timer = 120;
   bm_start_game_timer();
}

unsigned char bm_get_next_monster_start (signed int *pos_x, signed int *pos_y )
{
   unsigned char x, y;
   for ( x = 0; x < game->map_width; x++ )
   {
      for ( y = 0; y < game->map_height; y++ )
      {
         if ( game->map[y][x] == MAP_MONSTER)
         {
            *pos_x = x * 6;
            *pos_y = y * 6;
            game->map[y][x] = MAP_NONE;
            return 1;
         }
      }
   }
   return 0;
}

void bm_calculate_new_monsters_pos()
{
   unsigned char c;
   for(c=0; c < MAX_MONSTERS; c++)
      if(game->monsters[c].state & 0x03) //only lived monsters
      {
         game->monsters[c].state++;
         /* type 1 monsters */
         if(game->monsters[c].type == 1)
            game->monsters[c].state&=0x03;
         /* type 2 monsters - much fasters */
         else if(game->monsters[c].type == 2)
            game->monsters[c].state&=0x01;
         /* move monster? */
         if(game->monsters[c].state == 0)
         {
            game->monsters[c].state++;
            signed int x = game->monsters[c].x;
            signed int y = game->monsters[c].y;
            signed char map_x, map_y;
      
            switch(game->monsters[c].direction)
            {
               case 0x0:      //up
                  y-=PLAYER_STEP_VALUE;
                  break;
               case 0x1:      //left
                  x-=PLAYER_STEP_VALUE;
                  break;
               case 0x2:      //down
                  y+=PLAYER_STEP_VALUE;
                  break;
               case 0x3:      //right
                  x+=PLAYER_STEP_VALUE;
                  break;
            }
            //left upper monster's "corner"
            bm_pos_to_map( x, y, &map_x, &map_y );
            unsigned char wall = bm_check_wall(map_x, map_y);
            if(((wall != MAP_NONE)&&(wall != MAP_EXIT) && (wall != MAP_BONUS_POWER) && (wall != MAP_BONUS_LIFE) && (wall != MAP_BONUS_DRIVE) && (wall != MAP_BONUS_CASH) && (wall != MAP_BONUS_BOMB) )||(bm_check_bomb(map_x, map_y, NULL)))
            {
               bm_monster_change_direction(c);
               continue;
            }
            //left lower monster's "corner"
            bm_pos_to_map( x, y+5, &map_x, &map_y );
            wall = bm_check_wall(map_x, map_y);
            if(((wall != MAP_NONE)&&(wall != MAP_EXIT) && (wall != MAP_BONUS_POWER) && (wall != MAP_BONUS_LIFE) && (wall != MAP_BONUS_DRIVE) && (wall != MAP_BONUS_CASH) && (wall != MAP_BONUS_BOMB) )||(bm_check_bomb(map_x, map_y, NULL)))
            {
               bm_monster_change_direction(c);
               continue;
            }
            //right upper monster's "corner"
            bm_pos_to_map( x+5, y, &map_x, &map_y );
            wall = bm_check_wall(map_x, map_y);
            if(((wall != MAP_NONE)&&(wall != MAP_EXIT) && (wall != MAP_BONUS_POWER) && (wall != MAP_BONUS_LIFE) && (wall != MAP_BONUS_DRIVE) && (wall != MAP_BONUS_CASH) && (wall != MAP_BONUS_BOMB) )||(bm_check_bomb(map_x, map_y, NULL)))
            {
               bm_monster_change_direction(c);
               continue;
            }
            //right lower monster's "corner"
            bm_pos_to_map( x+5, y+5, &map_x, &map_y );
            wall = bm_check_wall(map_x, map_y);
            if(((wall != MAP_NONE)&&(wall != MAP_EXIT) && (wall != MAP_BONUS_POWER) && (wall != MAP_BONUS_LIFE) && (wall != MAP_BONUS_DRIVE) && (wall != MAP_BONUS_CASH) && (wall != MAP_BONUS_BOMB) )||(bm_check_bomb(map_x, map_y, NULL)))
            {
               bm_monster_change_direction(c);
               continue;
            }
            //checking player
/*            signed int pl_x = game->x;
            signed int pl_y = game->y;
            signed char pl_map_x, pl_map_y;
            bm_pos_to_map( pl_x, pl_y, &pl_map_x, &pl_map_y );
            if(map_x==pl_map_x && map_y==pl_map_y) */
            if(bm_distance(game->x, game->y, game->monsters[c].x, game->monsters[c].y)<5)
               bm_kill_player();
               
            game->monsters[c].x = x;
            game->monsters[c].y = y;
         }
      }
}

void bm_monster_change_direction(unsigned char num)
{
   unsigned char opposite_dir=rand()%5; //chance for change direction to other than
                                        //opposite is 1 to 5 (+1 to 4 from below random direction)
   switch ( game->monsters[num].direction)
   {
      case 0x0:      //up
         game->monsters[num].direction=(opposite_dir?2:rand()%4);
         break;
      case 0x1:      //left
         game->monsters[num].direction=(opposite_dir?3:rand()%4);
         break;
      case 0x2:      //down
         game->monsters[num].direction=(opposite_dir?0:rand()%4);
         break;
      case 0x3:      //right
         game->monsters[num].direction=(opposite_dir?1:rand()%4);
         break;
   }
}

void bm_kill_player()
{
   game->lives--;
   bm_play_sound(&tone_game_lost);

   if(!game->lives)
   {
      dprintf("kill player app.state = %d\n", app.state);
      game->gameover=1;
      bm_play_sound(&tone_gameover);
      #ifndef ENABLE_FADE
      send_message(MSG_REQUEST_QUIT);
      #endif
   }
   #ifndef ENABLE_FADE
   else
      bm_reload_level();
   #else
   game->fade_out = FADE_LENGTH;
   #endif
}

/* UI messages dispatcher */
extern rexx_uint game_dispatcher_id; /* dispatcher's call id */
rexx("game_dispatcher_id=add_5e0_id(game_dispatcher,locate('action_back_back'))"); /* auto creation of the id */

int game_dispatcher(int msg)
{
   switch(msg)
   {
      case MSG_D_INIT: // initialize game
         freegfx_open();
         app.old_lights_owner=lights_owner;
         lights_owner=LIGHTS_OWNER_ON;
         if(own_get_game_setting(GS_LIGHTS)) send_message(MSG_LIGHTS_ON);
         else send_message(MSG_LIGHTS_OFF);

         bm_start_timer();
         msg=MSG_R_STAY;
         break;

      case MSG_D_QUIT: // handle quit to menu
         bm_stop_timer();
         bm_stop_game_timer();
         lights_owner=app.old_lights_owner;
         send_message(MSG_TIMED_LIGHTS_ON);

         freegfx_close();

         msg=MSG_R_QUIT;
         break;

      case MSG_DISPLAY_MESSAGE: // quits to menu when message like "Charging" is displayed
         send_message(MSG_REQUEST_QUIT);
         break;

      case MSG_INCOMING_CALL: // quits to menu when someone calls
         lights_owner=app.old_lights_owner;
         send_message(MSG_TIMED_LIGHTS_ON);
   
         freegfx_close();
         break;

      case MSG_CALL_FINISHED: // comes after finishing the call
         msg=MSG_R_LEAVE;
         break;


      case MSG_KEYRELEASE:
         own_timer_abort(&key_timer);
         //only align
         //TODO: align player
         game->key=0xFF;
         break;

      case MSG_KEYPRESS:
            if (game->key==0xFF) //user press new key
            {
               game->key = own_get_unify_keycode(UKM_NPM5_CURSORS1);
               game->keycounter=0;
               //cheat :)
               game->cheatlen++;
               if(game->cheatlen<=5)
               {
                  game->cheatcode[game->cheatlen-1] = game->key;
                  dprintf("Current cheat code: %d %d %d %d %d\n", game->cheatcode[0], game->cheatcode[1], game->cheatcode[2], game->cheatcode[3], game->cheatcode[4]);
                  //add life
                  if (memcmp(game->cheatcode, cheat_life, 4) == 0)
                  {
                     game->lives++;
                     bm_play_sound(&tone_cheat);
                     game->key=0xFF;
                  }

                  //add bomb
                  if (memcmp(game->cheatcode, cheat_bomb, 4) == 0)
                  {
                     game->bomb_count++;
                     if(game->bomb_count > MAX_BOMBS)
                        game->bomb_count = MAX_BOMBS;
                     bm_play_sound(&tone_cheat);
                     game->key=0xFF;
                  }

                  //add power
                  if (memcmp(game->cheatcode, cheat_power, 5) == 0)
                  {
                     game->bomb_power++;
                     if(game->bomb_power > MAX_STRENGHT)
                        game->bomb_power = MAX_STRENGHT;
                     bm_play_sound(&tone_cheat);
                     game->key=0xFF;
                  }

                  //add manual driving
                  if (memcmp(game->cheatcode, cheat_drive, 5) == 0)
                  {
                     game->manual_driving = !game->manual_driving;
                     bm_play_sound(&tone_cheat);
                     game->key=0xFF;
                  }
               }
               if(game->key == KEY_HASH)
               {
                  game->cheatlen = 0;
                  memset(game->cheatcode, 0, 5);
               }
            }
            if((game->key==KEY_2)||(game->key==KEY_4)||(game->key==KEY_6)||(game->key==KEY_8))
            {
               if (game->keycounter==0)
                  own_timer_start(&key_timer,TT_MESSAGE,20);
               else
                  own_timer_start(&key_timer,TT_MESSAGE,10);
               game->keycounter++;
            }
            #ifdef ENABLE_FADE
            if(!(game->fade_in) && !(game -> fade_out))
            #endif
               bm_calculate_player_pos();
         break;
         
      case OMSG_TIMER:
         bm_start_timer();
         #ifdef ENABLE_FADE
         if(!(game->fade_in) && !(game -> fade_out))
         #endif
         {
            bm_calculate_fires();
            bm_calculate_bombs();
            bm_calculate_new_monsters_pos();
         }

         bm_draw();
         break;

      case OMSG_GAME_TIMER:
         bm_start_game_timer();
         game->timer--;
         if(!game->timer)
            bm_kill_player();
         msg = MSG_NOP;
         break;

      case MSG_KEYPRESS_CHAR:
         msg=MSG_NOP; // prevents our dispatcher from quiting when numeric keys are being pressed
         break;
   }
   return msg;
}


/* lists implementation TODO */
bm_list_t *bm_put(bm_list_t *item)
{
   if(item == NULL)
      return NULL;

   return NULL;
}
void *bm_readdata(bm_list_t *item);
bm_list_t *bm_remove(bm_list_t *item);



