/*****************************************************************************************
 *
 *  Copyright 2006+ by Cieszak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,

 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#ifndef GAME_H
#define GAME_H

#include <nokix/all.h>

/**************************************************************************************/

/* settings */

//#define DEBUG

#define MAX_MONSTERS 5
#define MAX_STRENGHT 5
#define MAX_BOMBS 5

#define BOMB_TIME 0x1F

#define MAX_MAP_HEIGHT      25
#define MAX_MAP_WIDTH       25


//do not use default map if no files found
//#define ONLY_MAP_FILES

#define ENABLE_FADE

#define FADE_LENGTH 0x08

#define PLAYER_STEP_VALUE 2

/* other definitions - DO NOT CHANGE*/

#define MAP_NONE        ' '
#define MAP_FIXED       'F'
#define MAP_BREAKEBLE   'f'
#define MAP_HIDDENEXIT  'e'
#define MAP_EXIT        'E'
#define MAP_HIDDENBONUS 'b'
#define MAP_BONUS_CASH  'C'
#define MAP_BONUS_POWER 'P'
#define MAP_BONUS_DRIVE 'D'
#define MAP_BONUS_LIFE  'L'
#define MAP_BONUS_BOMB  'B'
//used only in map preparation
#define MAP_PLAYER      'P'
#define MAP_MONSTER     'm'
#define MAP_RESERVED    'r'

/* direction */
#define DIR_NONE        0
#define DIR_LEFT        1
#define DIR_RIGHT       2
#define DIR_UP          3
#define DIR_DOWN        4

#define SCREEN_HEIGHT   (screen_info.height/6)
#define SCREEN_WIDTH    (screen_info.width/6)

#ifdef DEBUG
   #define dprintf printf
#else
   #define dprintf(...);
#endif


/* structs */

typedef struct
{
    unsigned char state;   //0: none; 1..3: live; 4..7: killed
    signed int x;
    signed int y;
    unsigned char direction;
    unsigned char type;       //1: normal, 2: 2x faster
} bm_monster_t;

typedef struct
{
    signed char map_x;
    signed char map_y;
    unsigned char state;   //0: none, 1..FIRETIME: fired
    unsigned char check;
} bm_fire_t;

typedef struct
{
   signed char map_x;
   signed char map_y;
   unsigned char state; // 0: none; 1..BOMB_TIME: used
} bm_bomb_t;

/* list implementation */

typedef struct list
{
   void *prev;
   void *next;
   void *data;
} bm_list_t;

struct game
{
   /* player */
   signed int x;
   signed int y;
   unsigned char check;
   unsigned char bomb_count;
   unsigned char bomb_power;
   unsigned char manual_driving;
   unsigned char timer;


   signed char lives;
   unsigned char level;

   unsigned int score;
   unsigned char killed;

   /* maps */
   TFILE** maps;  //all maps ids
   unsigned char maps_count;  //number of maps

   /* map */
   TFILE* map_id; //current map id
   char map_name[20];   //current map name
   unsigned char map_width;
   unsigned char map_height;
   char map[MAX_MAP_HEIGHT][MAX_MAP_WIDTH]; //current map

   /* game data */
   bm_bomb_t bombs[MAX_BOMBS];
   bm_monster_t monsters[MAX_MONSTERS];
   bm_fire_t fires[MAX_BOMBS*(MAX_STRENGHT*4+1)];
   
   /* game data - lists */
   bm_list_t *bombs_list;
   bm_list_t *monsters_list;
   bm_list_t *fires_list;

   /* other game infos */
   unsigned char key;
   unsigned char keycounter;
   unsigned char gameover;
   unsigned char error;

   #ifdef ENABLE_FADE
      signed char fade_in;
      signed char fade_out;
   #endif
   
   //cheats :)
   unsigned char cheatcode[5];
   unsigned char cheatlen;
};

extern struct game *game;

/**************************************************************************************/
//functions:

/* shows error msg and play error sound */
void error(char* text);

/* frees the game structure */
void bm_free_game(void);

/* checks player start position */
void bm_get_player_start(signed int *pos_x, signed int *pos_y);

/* calculate map coord */
void inline bm_pos_to_map(signed int pos_x, signed int pos_y, signed char *map_x, signed char *map_y);

/* calculate map coord */
void inline bm_map_to_pos(signed char map_x, signed char map_y, signed int *pos_x, signed int *pos_y);

// return what is on given coord
unsigned char bm_check_wall(signed char map_x, signed char map_y);

// return: 0 no bomb, 1 bomb
unsigned char bm_check_bomb(signed char map_x, signed char map_y, unsigned char *bomb_id);

// return: 0 no fire, 1 fire
unsigned char bm_check_fire(signed char map_x, signed char map_y);

// return: 0 no monster, 1 monster
unsigned char bm_check_monster(signed char map_x, signed char map_y, unsigned char *monster_id );

void bm_calculate_player_pos();

void bm_draw();

void bm_add_new_fires(signed char map_x, signed char map_y);

void bm_calculate_fires();

void bm_calculate_bombs();

void bm_start_timer();

void bm_stop_timer();

void bm_start_game_timer();

void bm_stop_game_timer();

void bm_play_sound(const ringtone_t *tone);

void bm_read_map();

void bm_next_level();

void bm_reload_level();

unsigned char bm_get_next_monster_start (signed int *pos_x, signed int *pos_y );

void bm_calculate_new_monsters_pos();

void bm_monster_change_direction(unsigned char num);

void bm_kill_player();

/* list implementation */
bm_list_t *bm_put(bm_list_t *item);
void *bm_readdata(bm_list_t *item);
bm_list_t *bm_remove(bm_list_t *item);



#endif

