/*****************************************************************************************
 *
 *  Copyright 2006+ by Cieszak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#include <nokix/all.h>

/**************************************************************************************/

extern const struct bitmap bmap_intro;          // intro bitmap
extern rexx_uint menu_dispatcher_id;            // dispatcher id of the menu

static const unsigned intro_timer=MSG_SELECT;   // message to be sent automaticaly after 3s

/**************************************************************************************/

/* UI messages dispatcher */
extern rexx_uint intro_dispatcher_id; /* dispatcher's call id */
rexx("intro_dispatcher_id=add_5e0_id(intro_dispatcher)"); /* auto creation of the id */

unsigned intro_dispatcher(unsigned msg)
{
    switch(msg)
    {
        case MSG_D_INIT:
            // open screen for intro bitmap
            freegfx_open();

            // start the drawing operation
            freegfx_begin(true);
   
            // draw the bitmap
            own_draw_bitmap(&bmap_intro,0,0,DM_SET);
   
            // end the drawing operation
            freegfx_end(false);

            // start the timer to close the intro (by sending MSG_SELECT) after 3 seconds
            own_timer_start(&intro_timer,TT_MESSAGE,3*TICKS_PER_SECOND);
   
            // stay in the dispatcher
            msg=MSG_R_STAY;
            break;

        case MSG_D_QUIT_NOW:
        case MSG_D_QUIT:
            // stop the timer
            own_timer_abort(&intro_timer);

            // close the screen
            freegfx_close();

            // confirm exit
            msg=MSG_R_QUIT;
            break;

        case MSG_KEY_UP:
        case MSG_KEY_DOWN:
        case MSG_SELECT: // either main Navi key was pressed or timer reached zero
            // stop the timer
            own_timer_abort(&intro_timer);

            // close the screen
            freegfx_close();

            // start the menu dispatcher
            send_message(MSG_START_HANDLER|MSG_1P,menu_dispatcher_id);
            break;
    }
    return msg;
}

/**************************************************************************************/

