/*****************************************************************************************
 *
 *  Copyright 2006+ by Cieszak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ****************************************************************************************/

#include "menu.h"
#include "game.h"

#define OMSG_START_GAME     	(0x1f00)    // message sent by our "New game" and "Continue" menu items (see _project.nrx)
#define OMSG_INSTRUCTIONS   	(0x1f01)    // message sent by our "Instructions" menu item (see _project.nrx)
#define OMSG_HISCORE    	   (0x1f02)    // message sent by our "Instructions" menu item (see _project.nrx)

/**************************************************************************************/

/* expose preprocessor constants to Rexx */
REXX_EXPOSE(OMSG_START_GAME);
REXX_EXPOSE(OMSG_INSTRUCTIONS);
REXX_EXPOSE(OMSG_HISCORE);

/**************************************************************************************/

extern rexx_uint game_dispatcher_id;               // dispatcher id of the game
extern rexx_uint hiscoreid;                        //id of hiscore eeprom item

extern const struct menu_item i_continue;          // menu items
extern const struct menu_item i_new_game;
extern const struct menu_item i_instructions;
extern const struct menu_item i_top_score;
extern const struct menu_item i_clear_top_score;

extern const nokstr_t str_game_name;
extern const nokstr_t str_instr_text;
extern const nokstr_t str_top_score_cleared;

/**************************************************************************************/

static void make_game_menu(void)
{
   while(menu_get_item(1,1)) menu_rem_item(1,1);

   if(game) menu_add_item(1,0,&i_continue);
   menu_add_item(1,0,&i_new_game);
   menu_add_item(1,0,&i_instructions);
   if(app.hiscore > 0)
   {
      menu_add_item(1,0,&i_top_score);
      menu_add_item(1,0,&i_clear_top_score);
   }
}

/**************************************************************************************/

/* UI messages dispatcher */
extern rexx_uint menu_dispatcher_id; /* dispatcher's call id */
rexx("menu_dispatcher_id=add_5e0_id(menu_dispatcher)"); /* auto creation of the id */

unsigned menu_dispatcher(unsigned msg)
{
   switch(msg)
   {
      case MSG_D_INIT: // initialize and open menu
         /* read hiscore */
         dprintf("hiscoreid = %d\n", hiscoreid);
         if(hiscoreid != -1)
         {
            eeprom_read(&app.hiscore, hiscoreid);
         }
         else
         {
            app.hiscore = 0;
            if(!TFS_DIRECTORY(tfsstat("/.config")))
            {
               if(TFS_FILE(tfsstat("/.config")))
               {
                  dprintf(".config is a file, not a directory :/\n");
               }
               else
                  tfserror(tfsmkdir("/.config", TFS_A_HIDDEN));
            }
            int fd = tfsopen("/.config/bomberman.hi", TFS_RDONLY);
            if(fd > 0)
            {
               tfsread(fd, &app.hiscore, sizeof(app.hiscore));
               tfsclose(fd);
            }
         }
         /* main menu */
         menu_create(1,0x60,1,1,&str_game_name);
         make_game_menu();
         menu_open(1,1);

         app.state=STATE_MENU;

         msg=MSG_R_STAY;
         break;

      case MSG_D_QUIT: // handle menu quit
         menu_delete(1);

         msg=MSG_R_QUIT;
         break;

      case MSG_D_QUIT_NOW: // called when we get back from some other dispatcher
         {
            unsigned char oldstate=app.state;

            app.state=STATE_MENU;

            if(oldstate==STATE_GAME) // if we got back from game dispatcher
            {
               bm_stop_timer();
               bm_stop_game_timer();
               if(game->gameover)
               {
                  if(game->score > app.hiscore)
                  {
                     bm_play_sound(&tone_game_won);
                     app.hiscore = game->score;
                     if(hiscoreid != -1)
                     {
                        eeprom_write(&app.hiscore, hiscoreid);
                     }
                     else
                     {
                        if(!TFS_DIRECTORY(tfsstat("/.config")))
                        {
                           if(TFS_FILE(tfsstat("/.config")))
                           {
                              dprintf(".config is a file, not a directory :/\n");
                           }
                           else
                              tfserror(tfsmkdir("/.config", TFS_A_HIDDEN));
                        }
                        int fd = tfsopen("/.config/bomberman.hi", TFS_APPEND|TFS_CREATE);
                        if(fd > 0)
                        {
                           tfserror(tfsseek(fd, 0, TFS_BEGIN));
                           tfserror(tfswrite(fd, &app.hiscore, sizeof(app.hiscore)));
                           tfserror(tfsclose(fd));
                        }
                        else
                           tfserror(fd);
                     }
                     send_message(MSG_DISPLAY_MESSAGE|MSG_2P, DMI_TOP_SCORE, w_sprintf(&str_top_score_n, app.hiscore));
                  }
                  bm_free_game();

               }

            }

            make_game_menu();
            menu_reopen();

            msg=MSG_R_STAY;
            break;
         }

      case OMSG_START_GAME: // "New game" or "Continue" was selected from menu
         if(!game)
         {
            game=malloc(sizeof(struct game));
            memset(game,0,sizeof(struct game));
            srand(get_timestamp());
            memset(game->cheatcode, 0, 5);
            //cheat disabled
            game->cheatlen = 5;
         }
   
         /* since OMSG_START_GAME message is sent by both "New game" and "Continue"
            items, we have to distinguish between them and initialize the game to
            start values only when "New game" was selected;
            we do it by checking the msg_argv[0] of the message (it was set in _project.nrx) */
         if(msg_argv[0])
         {
            //game continue handling
            bm_start_game_timer();
         }
         else
         {
            memset(game,0,sizeof(struct game));
            game->bombs_list = NULL;
            game->monsters_list = NULL;
            game->fires_list = NULL;

            dprintf("game %p\n", game);
            TFILE *fp=NULL;
            game->maps_count=0;
            while((fp=tfsnext(fp, "/bomberman")))
            {
               if (TFS_DIRECTORY(fp)) continue; //directory
               game->maps=(TFILE **)realloc(game->maps,(game->maps_count+1)*sizeof(TFILE *));
               game->maps[game->maps_count]=fp;
               game->maps_count++;
            }

            game->lives=3;
            game->bomb_power=1;
            game->bomb_count=1;
            bm_next_level();
         }

         #ifdef ONLY_MAP_FILES
         if(!game->maps_count)
         {
            free_game();
            display_msg(DMI_ERROR, &str_no_levels);
            send_message(MSG_REQUEST_QUIT_OK);
         }
         else
         #endif
         {
            // start the game dispatcher
            send_message(MSG_START_HANDLER|MSG_1P,game_dispatcher_id); /* start the game */
            app.state=STATE_GAME;
         }

         msg=MSG_NOP;
         break;

      case OMSG_INSTRUCTIONS: // "Instructions" was selected from menu
           // displays the instructions screen
         send_message(MSG_START_HANDLER|MSG_2P,id_game_instructions,str_instr_text);

         app.state=STATE_INSTRUCTIONS;

         msg=MSG_NOP;
         break;
         
      case OMSG_HISCORE: // HiScore or ClearHiScore
         if(msg_argv[0] == 0) //show hiscore
         {
            send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_TOP_SCORE,w_sprintf(&str_top_score_n, app.hiscore));
            menu_reopen();
         }
         else //clear hiscore
         {
            dialog_open((char *)&str_are_you_sure,&dialog_bool);
            app.state=STATE_CLEAR_DLG;
         }
         msg = MSG_NOP;
         break;

      case MSG_DIALOG_ACCEPTED:
         if(app.state == STATE_CLEAR_DLG)
         {
            //clearing top score
            app.hiscore = 0;
            if (hiscoreid != -1)
            {
               eeprom_write(&app.hiscore, hiscoreid);
            }
            else
            {
               if(!TFS_DIRECTORY(tfsstat("/.config")))
               {
                  if(TFS_FILE(tfsstat("/.config")))
                  {
                     dprintf(".config is a file, not a directory :/\n");
                  }
                  else
                     tfserror(tfsmkdir("/.config", TFS_A_HIDDEN));
               }
               int fd = tfsopen("/.config/bomberman.hi", TFS_APPEND|TFS_CREATE);
               tfsseek(fd, 0, TFS_BEGIN);
               tfswrite(fd, &app.hiscore, sizeof(app.hiscore));
               tfsclose(fd);
            }
            send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_INFORMATION,&str_top_score_cleared);
            make_game_menu();
            menu_reopen();
            app.state=STATE_MENU;
         }
         msg = MSG_NOP;
         break;

   }
   return msg;
}

/**************************************************************************************/

